/*
 * Renderer.h
 *
 * Copyright (C) 2009 Matthias Schiffer <matthias@gamezock.de>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ZOOM_RENDERER_H_
#define ZOOM_RENDERER_H_

#include "gl.h"
#include "BSPTree.h"
#include "Shader.h"
#include "ShadowVolume.h"
#include <limits>


namespace Zoom {

class Renderer {
  public:
    Renderer();

    //void render(const BSPTree &tree);

    template <typename T>
    void render(const T &triangles, const vmml::vec3f &lightPos) {
      if(triangles.empty())
        return;

      // Create shadow volumes
      std::vector<ShadowVolume> shadowVolumes;
      for(typename T::const_iterator t = triangles.begin(); t != triangles.end(); ++t) {
        shadowVolumes.push_back(ShadowVolume(t->triangle, lightPos));
      }


      glLightfv(GL_LIGHT0, GL_POSITION, vmml::vec4f(lightPos, 1).array);

      // Render with ambient light
      Shader::enable(ambientShader);
      glBlendFunc(GL_ONE, GL_ZERO);
      glDepthFunc(GL_LEQUAL);
      glStencilFunc(GL_ALWAYS, 0, std::numeric_limits<GLuint>::max());
      glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

      useTexture(triangles.front().triangle.getTexture());

      glBegin(GL_TRIANGLES);
      for(typename T::const_iterator t = triangles.begin(); t != triangles.end(); ++t) {
        renderTriangle(t->triangle);
      }
      glEnd();

      // Render shadow volumes
      glClear(GL_STENCIL_BUFFER_BIT);
      glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
      glDepthMask(GL_FALSE);
      Shader::disable();

      glFrontFace(GL_CCW);
      glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);

      glBegin(GL_TRIANGLES);
      for(std::vector<ShadowVolume>::iterator v = shadowVolumes.begin(); v != shadowVolumes.end(); ++v) {
        v->render();
      }
      glEnd();

      glFrontFace(GL_CW);
      glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);

      glBegin(GL_TRIANGLES);
      for(std::vector<ShadowVolume>::iterator v = shadowVolumes.begin(); v != shadowVolumes.end(); ++v) {
        v->render();
      }
      glEnd();

      glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
      glDepthMask(GL_TRUE);
      glFrontFace(GL_CCW);

      // Render with point light
      Shader::enable(lightShader);
      glBlendFunc(GL_ONE, GL_ONE);
      glDepthFunc(GL_EQUAL);
      glStencilFunc(GL_EQUAL, 0, std::numeric_limits<GLuint>::max());
      glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

      useTexture(triangles.front().triangle.getTexture());

      glBegin(GL_TRIANGLES);
      for(typename T::const_iterator t = triangles.begin(); t != triangles.end(); ++t) {
        renderTriangle(t->triangle);
      }
      glEnd();
    }

    struct TextureSorter {
      bool operator() (const BSPTree::TriangleRecord &t1, const BSPTree::TriangleRecord &t2) {
        return (t1.triangle.getTexture() < t2.triangle.getTexture());
      }
    };

  private:
    void renderTriangle(const Triangle &t);
    void useTexture(unsigned texture);

    class RenderVisitor {
      public:
        RenderVisitor(Renderer *renderer0) : renderer(renderer0) {}

        void operator() (const BSPTree::TriangleRecord &t) const {
          renderer->renderTriangle(t.triangle);
        }

      private:
        Renderer *renderer;
    };

    boost::shared_ptr<Shader> ambientShader, lightShader;
    unsigned activeTexture;

    const RenderVisitor renderVisitor;
};

}

#endif /* ZOOM_RENDERER_H_ */
